#!/usr/bin/env python
import argparse
import os
import shutil
import subprocess
import tempfile
from typing import Iterable

from Bio import SeqIO
from Bio import SeqRecord


# Reading the FASTA file and filtering sequences
def filter_sequences(
    input_fasta: str, names_to_keep: list[str]
) -> Iterable[SeqRecord.SeqRecord]:
    for record in SeqIO.parse(input_fasta, "fasta"):
        if record.id in names_to_keep:
            yield record


# Function to run MAFFT
def run_mafft(input_fasta: str):
    # MAFFT command
    mafft_cmd = ["mafft", "--auto", input_fasta]
    return subprocess.run(mafft_cmd, capture_output=True, text=True)


def run_trimal(input_aln: str, output_phy: str, tree_method: str = "raxml"):
    # trimal command
    if tree_method == "raxml":
        trimal_cmd = [
            "trimal",
            "-in",
            input_aln,
            "-phylip",
            "-out",
            output_phy,
        ]
    else:
        # output fasta
        trimal_cmd = [
            "trimal",
            "-in",
            input_aln,
            "-out",
            output_phy,
        ]
    return subprocess.run(trimal_cmd, capture_output=True, text=True)


# Function to run RAxML
def run_raxml(alignment_file: str, output_dir: str):
    # RAxML command
    raxml_cmd = [
        "raxmlHPC-PTHREADS-SSE3",
        "-s",
        alignment_file,
        "-m",
        "GTRGAMMA",
        "-f",
        "a",
        "-p",
        "42",
        "-x",
        "60",
        "-N",
        "100",
        "-n",
        "whatever",
        "-T",
        str(16),
        "-w",
        output_dir,
    ]
    return subprocess.run(raxml_cmd, capture_output=True, text=True, cwd=".")


def run_fasttree(alignment_file: str, output_tree: str):
    """Run FastTre on nucleotide alignment file and save the tree by taking input from
    stdin and redirecting stdout to a file.
    """
    # FastTree command
    fasttree_cmd = [
        "FastTree",
        "-nt",
        "-gtr",
    ]
    print(f"Running FastTree with command: {' '.join(fasttree_cmd)}")
    # return subprocess.run(fasttree_cmd, capture_output=True, text=True, cwd=".")
    with open(alignment_file, "r") as aln, open(output_tree, "w") as tree:
        return subprocess.run(fasttree_cmd, stdin=aln, stdout=tree, text=True, cwd=".")


def process_sequences(
    zotu_fasta_path: str,
    alignment_out: str,
    tree_out: str,
    seq_names: list[str] | None = None,
    tree_method: str = "raxml",
) -> int:
    if seq_names is None:
        input_fasta = zotu_fasta_path
    else:
        # with tempfile.NamedTemporaryFile(
        #     mode="w", delete=False, suffix=".fasta"
        # ) as temp_fasta:
        temp_fasta = tempfile.NamedTemporaryFile(mode="w", suffix=".fasta")
        # Filter and write sequences to a temp file
        _ = SeqIO.write(
            list(filter_sequences(zotu_fasta_path, seq_names)),
            temp_fasta,
            "fasta",
        )
        input_fasta = temp_fasta.name

    # Run MAFFT on the temp file and check if it ran successfully
    mafft_result = run_mafft(input_fasta)
    if mafft_result.returncode != 0:
        print("Error in running MAFFT:", mafft_result.stderr)
        os.remove(temp_fasta.name)
        return 1
    temp_aln = tempfile.NamedTemporaryFile(mode="w", suffix=".aln")
    temp_aln.write(mafft_result.stdout)
    temp_aln.flush()

    # Run trimal on the alignment file and check if it ran successfully
    trimal_result = run_trimal(temp_aln.name, alignment_out, tree_method)
    temp_aln.close()
    if trimal_result.returncode != 0:
        print("Error in running trimal:", trimal_result.stderr)
        return 1
    else:
        print(
            "MAFFT and trimal completed successfully. Alignment saved as:",
            alignment_out,
        )

    if tree_method == "raxml":
        # Run RAxML, output to a temporary directory, copy the tree file to the output path
        with tempfile.TemporaryDirectory() as temp_dir:
            raxml_result = run_raxml(alignment_out, temp_dir)
            tree_file = os.path.join(temp_dir, "RAxML_bestTree.whatever")
        # Check if RAxML ran successfully
        if raxml_result.returncode != 0:
            print("Error in running RAxML:", raxml_result.stderr)
            return 1
        else:
            print(
                "RAxML completed successfully. Tree saved in Newick format as:",
                tree_out,
            )
            shutil.copy(tree_file, tree_out)
    elif tree_method == "fasttree":
        fasttree_result = run_fasttree(alignment_out, tree_out)
        if fasttree_result.returncode != 0:
            print("Error in running FastTree:", fasttree_result.stderr)
            return 1
        else:
            print(
                "FastTree completed successfully. Tree saved in Newick format as:",
                tree_out,
            )
    else:
        raise ValueError("Invalid tree method:", tree_method)

    if seq_names is not None:
        temp_fasta.close()

    return 0


def main() -> None:
    # if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate a phylogenetic tree from a FASTA file of sequences and a list of sequence names"
    )
    _ = parser.add_argument(
        "-i",
        "--input_fasta",
        help="Input FASTA file",
        required=True,
        type=str,
        metavar="FASTA",
    )
    _ = parser.add_argument(
        "-m",
        "--tree_method",
        help="Method to generate the tree",
        required=False,
        type=str,
        default="fasttree",
        choices=["raxml", "fasttree"],
    )
    args = parser.parse_args()

    # Read the sequence names file

    # Process the sequences
    # prefix is dirname + anything before first dot
    prefix = os.path.join(
        os.path.dirname(args.input_fasta),
        os.path.basename(args.input_fasta).split(".")[0],
    )
    aln_file = prefix + (".phy" if args.tree_method == "raxml" else ".alnfna")
    tree_file = prefix + ".newick"
    process_sequences(args.input_fasta, aln_file, tree_file, tree_method="fasttree")

    # Read the tree file and print the tree
    # tree = Phylo.read(args.output, "newick")
    # Phylo.draw_ascii(tree)



import pytest
from pathlib import Path


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


@pytest.mark.parametrize("amplicon_type", ["16s"])
@pytest.mark.parametrize("level", ["genus"])
@pytest.mark.parametrize(
    "distance_method, log10, annotated, suffix",
    [
        ("euclid", True, False, ""),  # Euclidean with log10, no annotation
        ("braycurtis", False, False, ""),  # Bray-Curtis without log10, no annotation
        ("euclid", True, True, "annot_"),  # Euclidean with log10 and annotation
        ("braycurtis", False, True, "annot_"),  # Bray-Curtis with annotation
    ],
)
def test_plot_beta_diversity(
    script_runner,
    test_data_dir,
    amplicon_type,
    level,
    distance_method,
    log10,
    annotated,
    suffix,
):
    """
    Test script for plotting two kinds of beta diversity (with or without annotation)
    """

    # Define test variables
    output_unoise3_dir = "output_unoise3_16s"
    fig_dir = test_data_dir / "figs"

    # Prepare input and output directories
    input_file = test_data_dir / output_unoise3_dir / "unoise3_zotu.biom"
    metadata_file = test_data_dir / "metadata" / f"bulk_{amplicon_type}.tsv"
    taxonomy_file = (
        test_data_dir / output_unoise3_dir / "unoise3_zotu_rrndb_processed.tsv"
    )
    output_dir = fig_dir / f"beta_diversity_{suffix}{amplicon_type}"

    # Construct the base command
    cmd = [
        "plot_dm.py",
        "-i",
        str(input_file),
        "-m",
        str(metadata_file),
        "-t",
        str(taxonomy_file),
        "-f",
        output_dir,
        "-l",
        "otu",
        level,
        "-d",
        distance_method,
        "--hue",
        "rep_group",
    ]

    # Add log10 and transformation if applicable
    if log10:
        cmd.extend(["--log10", "-tr", "relative"])

    # Add annotation option if applicable
    if annotated:
        cmd.append("-a")

    # Run the script
    ret = script_runner.run(cmd)

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"



import pytest
from pathlib import Path


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


def test_rdp(script_runner, test_data_dir):
    """
    run_rdp_classifier.py \
        -i $working_dir/$output_unoise3_dir/unoise3_zotu.fa \
        -d $rdp_db
    """
    # Prepare input and output directories
    input_file = test_data_dir / "output_unoise3_16s/unoise3_zotu.fa"
    db_file = "16srrna"

    # Run the script
    ret = script_runner.run(
        [
            "run_rdp_classifier.py",
            "-i",
            str(input_file),
            "-d",
            str(db_file),
        ]
    )

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    # Optionally, add more assertions to verify the content of the output file


#!/usr/bin/env python
import argparse
import subprocess
import os

import yaml

import masato


# RDP_CLASSIFIER_PATH: str = masato.__path__[0] + "/../../dist/rdp_classifier_2.14/dist/rdp_classifier"
# INFO_PATH = os.path.join(os.path.dirname(__file__), "setup/info.yaml")
# with open(INFO_PATH) as f:
#     info = yaml.safe_load(f)
#     rdp_classifier_path = info["rdp_classifier_path"]
#     if rdp_classifier_path is None:
#         RDP_CLASSIFIER_PATH = (
#             masato.__path__[0]
#             + "/../../dist/rdp_classifier_2.14/dist/rdp_classifier"
#         )
#     else:
#         RDP_CLASSIFIER_PATH = rdp_classifier_path
RDP_CLASSIFIER_PATH = (
    masato.__path__[0] + "/../../dist/rdp_classifier_2.14/dist/rdp_classifier"
)


def main():
    """
    Run these two commands with arguments for input zotu fasta and database. Output paths are inferred.

    rdp_classifier \
        $working_dir/$output_unoise3_dir/unoise3_zotu.fa \
        -d $rdp_db \
        -o $working_dir/$output_unoise3_dir/unoise3_zotu_rrndb.tsv \
        -h $working_dir/$output_unoise3_dir/unoise3_zotu_rrndb.hier.tsv

    $src_dir/process_rrndb.py \
        -i $working_dir/$output_unoise3_dir/unoise3_zotu_rrndb.tsv \
        -o $working_dir/$output_unoise3_dir/unoise3_zotu_rrndb_processed.tsv
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str)
    parser.add_argument("-d", "--database", type=str)

    args = parser.parse_args()
    input_path = args.input
    database_path = args.database
    base_name = os.path.splitext(input_path)[0]
    raw_output_path = f"{base_name}_rrndb_raw.tsv"
    raw_hier_path = f"{base_name}_rrndb_raw_hier.tsv"
    output_path = f"{base_name}_rrndb_processed.tsv"

    # Run the commands
    ret = subprocess.run(
        [
            RDP_CLASSIFIER_PATH,
            input_path,
            "-d",
            database_path,
            "-o",
            raw_output_path,
            "-h",
            raw_hier_path,
        ]
    )
    assert (
        ret.returncode == 0
    ), f"rdp_classifier failed with return code {ret.returncode}"

    ret = subprocess.run(["process_rrndb.py", "-i", raw_output_path, "-o", output_path])
    assert (
        ret.returncode == 0
    ), f"process_rrndb.py failed with return code {ret.returncode}"



import os
import queue
import tempfile
import subprocess
from itertools import cycle

import numpy as np
import pandas as pd
import dendropy
import matplotlib.pyplot as plt
from tqdm.auto import tqdm


def get_coords(
    node: dendropy.Node, max_x: float, overwrite: bool = False
) -> tuple[float, float]:
    if not hasattr(node, "attributes"):
        node.attributes = {}
    attributes = node.attributes
    x = attributes.get("x")
    y = attributes.get("y")
    plot = 1
    if not overwrite and x is not None and y is not None:
        return x, y
    if x is None or overwrite:
        x = max_x - node.distance_from_root()
    if y is None or overwrite:
        if node.is_leaf():
            if y is None:
                raise ValueError(
                    f"Leaf node {node} has no y value. Set leaf y value first."
                )
        elif node.num_child_nodes() == 0:
            y = -999
            plot = 0
        else:
            y = (
                sum(
                    [
                        get_coords(child, max_x=max_x, overwrite=overwrite)[1]
                        for child in node.child_node_iter()
                    ]
                )
                / node.num_child_nodes()
            )
    attributes.update({"x": x, "y": y, "plot": plot})
    return x, y


def _get_node_label(node: dendropy.Node) -> str:
    return node.taxon.label if node.taxon else node.label


def get_taxonomy_tree(
    taxon_df: pd.DataFrame, default_branch_length: float = 1.0
) -> dendropy.Tree:
    with (
        tempfile.NamedTemporaryFile(mode="w", suffix=".tsv") as temp_tsv,
        tempfile.TemporaryDirectory() as temp_dir,
    ):
        taxon_df.iloc[:, ::2].apply(lambda x: ";".join(x), axis=1).to_csv(
            temp_tsv.name, header=None, sep="\t"
        )
        proc = subprocess.run(
            [
                "gappa",
                "prepare",
                "taxonomy-tree",
                "--keep-singleton-inner-nodes",
                "--keep-inner-node-names",
                # "--replace-invalid-chars",
                "--allow-file-overwriting",
                "--file-prefix",
                "whatever_",
                "--taxon-list-file",
                temp_tsv.name,
            ],
            cwd=temp_dir,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if proc.returncode != 0:
            raise ValueError("Error in running gappa prepare taxonomy-tree.")

        tree = dendropy.Tree.get(
            path=os.path.join(temp_dir, "whatever_taxonomy_tree.newick"),
            schema="newick",
            rooting="force-rooted",
            preserve_underscores=True,
        ).extract_tree(
            node_filter_fn=lambda n: n.taxon.label in taxon_df.index,
            is_apply_filter_to_leaf_nodes=True,
            suppress_unifurcations=False,
        )
        for edge in tree.postorder_edge_iter():
            if edge.length is None:
                edge.length = default_branch_length

    return tree


def _calc_y(
    tree: dendropy.Tree, taxon_df: pd.DataFrame, collapse_level: str, base: float = 0
) -> dict[str, float]:
    taxon_zotu_count = taxon_df[collapse_level].value_counts().to_dict()
    zotu2taxon = taxon_df[collapse_level].to_dict()
    taxon2y = {}
    taxon2ys = {
        t: np.linspace(-0.4, 0.4, c) if c > 1 else [0]
        for t, c in taxon_zotu_count.items()
    }
    taxon_count_now = {}
    stack = [tree.seed_node]
    y = base
    ordered_taxa = []
    while stack:  # depth-first search
        node = stack.pop()
        node_name = _get_node_label(node)
        if node_name in taxon_zotu_count:
            taxon2y[node_name] = y
            taxon_count_now[node_name] = 0
            node.attributes = {"y": y}
            y += 1
            ordered_taxa.append(node_name)
        if node.is_leaf():
            node_taxon = zotu2taxon[node_name]
            node.attributes = {
                "y": taxon2y[node_taxon]
                + taxon2ys[node_taxon][taxon_count_now[node_taxon]]
            }
            taxon_count_now[node_taxon] += 1
        else:
            child_nodes = np.array(node.child_nodes(), dtype=object)
            names = np.array(
                [name for n in child_nodes if (name := _get_node_label(n)) is not None]
            )
            order = np.argsort(names)
            stack.extend(child_nodes[order][::-1].tolist())
    return ordered_taxa


def get_taxon2color(
    taxon_df: pd.DataFrame,
    levels_of_interest: list[str] | None = None,
    target_level: str | None = None,
) -> dict[str, str]:
    """tab20c for genus, tab20b for family, set3 for order, dark2 for class, set1 for phylum.
    If target_level is not None, only levels above the target level will be colored.
    """
    levels_all = ["phylum", "class", "order", "family", "genus", "species"]
    if levels_of_interest is None:
        levels = levels_all
        cmaps = ["tab20b", "Set1", "tab20c", "Set3", "Dark2"]
    else:
        levels = levels_of_interest
        cmaps = ["tab20b", "tab20c", "Dark2", "Set1", "Set3"]

    # target_level_idx = 999
    # try:
    #     target_level_idx = levels.index(target_level)
    # except ValueError:
    #     pass

    # colors should be circular in case there are more taxon than colors
    colors = list(map(lambda c: cycle(plt.get_cmap(c).colors), cmaps))
    level2color = {l: c for l, c in zip(levels, colors)}
    res = {}
    # for level, color in enumerate(level2color.items()):
    for level in levels:
        if levels_all.index(level) < levels_all.index(target_level):
            color = level2color[level]
            for taxon in taxon_df[level].unique():
                res[taxon] = next(color)
        else:
            for taxon in taxon_df[level].unique():
                res[taxon] = None
    res["unknown"] = None
    return res


def get_taxon2marker(
    taxon_df: pd.DataFrame, levels_of_interest: list[str] | None = None
) -> dict[str, str]:
    # hexagon for domain, pentagon for phylum, square for class, triangle for order, star for family, circle for genus
    markers = "*hps^o"
    levels = ["domain", "phylum", "class", "order", "family", "genus"]
    if levels_of_interest is None:
        levels_of_interest = levels
    level2marker = {l: m for l, m in zip(levels, markers) if l in levels_of_interest}
    level2marker["species"] = None  # no marker for species, but do add it to the dict
    res = {}
    for level, marker in level2marker.items():
        for taxon in taxon_df[level].unique():
            res[taxon] = marker
    res["unknown"] = None
    return res


def plot_tree(
    tree: dendropy.Tree,
    node_label2marker: dict[str, float],
    node_label2color: dict[str, str],
    node_label2alpha: dict[str, float],
    ax: plt.Axes,
    color_propagate: bool = False,
    annotate: bool = False,
    nodes_to_drop: set[str] = set(),
    terminal_nodes=None,
) -> None:
    """Plot dendrogram with matplotlib.
    Plotting starts from root to leaf, and each iteration draws a horizontal line from parent
        coordinates to the x coordinate of its children, and a vertical line that spans
        the y coordinates of its children. Along the way, make sure that
        each child has the same x coordinate, otherwise the tree is ill-formed.
    """
    node_label2color = dict(node_label2color)
    # terminal_nodes = set(terminal_nodes)
    # nodes_to_plot = queue.Queue()
    # nodes_to_plot.put(tree.seed_node)
    # while not nodes_to_plot.empty():
    # node = nodes_to_plot.get()
    nodes_to_plot = [tree.seed_node]
    while nodes_to_plot:
        node = nodes_to_plot.pop()
        node_name = _get_node_label(node)
        if not node.attributes.get("plot") or node_name in terminal_nodes:
            continue
        x, y = get_coords(node, max_x=None)

        # if node.parent_node is not None:
        #     xp, yp = get_coords(node.parent_node, max_x=None)
        #     # horizontal line from parent to child
        #     ax.plot(
        #         [xp, x],
        #         [y, y],
        #         "k",
        #         alpha=edge_label2alpha.get(node.edge.label, 1),
        #         solid_capstyle="round",
        #         lw=1,
        #     )
        #     if x == 0:
        #         plotted_nodes.append(node)
        # x coord of child nodes

        color = node_label2color.get(node_name, "k")
        child_nodes = np.array(node.child_nodes(), dtype=object)

        if len(child_nodes):
            child_coords = np.array(
                [
                    get_coords(child, max_x=None)
                    for child in child_nodes
                    if child.attributes.get("plot")
                ]
            )
            order = child_coords[:, 1].argsort()
            child_nodes = child_nodes[order]
            child_coords = child_coords[order]

            if len(child_coords) > 1:
                # plot a vertical line from the y coord of the first child to the last child
                if len(np.unique(child_coords[:, 0])) > 1:
                    # validate that all child nodes have the same x coord
                    raise ValueError(
                        f"Child nodes of {node} have different x coordinates."
                    )
            for i, (node_child, (xc, yc)) in enumerate(zip(child_nodes, child_coords)):
                child_name = _get_node_label(node_child)
                if child_name in nodes_to_drop:
                    continue
                # if yc > y or len(child_coords) == 1:
                #     yt = max(y, child_coords[i - 1, 1]) if i > 0 else y
                # else:
                #     yt = (
                #         min(y, child_coords[i + 1, 1])
                #         if i < len(child_coords) - 1
                #         else y
                #     )
                yt = (
                    max(y, child_coords[i - 1, 1])
                    if yc > y or len(child_coords) == 1
                    else min(y, child_coords[i + 1, 1])
                )

                # alpha goes with child node, color goes with parent node
                ax.plot(
                    [x, x, xc],
                    [yt, yc, yc],
                    alpha=node_label2alpha.get(child_name, 1),
                    color=color,
                    lw=2,
                    solid_capstyle="round",
                    solid_joinstyle="round",
                )
                if color_propagate and (
                    not node_label2color.get(child_name)
                    or not node_label2marker.get(child_name)
                ):
                    node_label2color[child_name] = color
                nodes_to_plot.append(node_child)
        marker = node_label2marker.get(node_name)
        if marker is not None:
            ax.scatter(
                x,
                y,
                s=5,
                color=color,
                alpha=node_label2alpha.get(node_name, 1),
                marker=marker,
                zorder=10,
            )
        if annotate:
            ax.text(
                x,
                y,
                node_name,
                fontsize=3,
                ha="right",
                va="center",
                color="k",
                zorder=10,
            )



import pytest
from pathlib import Path


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


@pytest.mark.parametrize("amplicon_type, prefix", [("16s", "16S-U")])
def test_unoise3(script_runner, test_data_dir, amplicon_type, prefix):
    """
    Test for running the unoise3 command.
    """
    # Prepare input and output directories
    output_unoise3_dir = f"output_unoise3_{amplicon_type}"
    input_file = test_data_dir / output_unoise3_dir / "merged.fq.gz"
    output_file = test_data_dir / output_unoise3_dir / "unoise3_zotu.fa"

    # Run the script
    ret = script_runner.run(
        [
            "usearch_workflow.py",
            "unoise3",
            "-i",
            str(input_file),
            "-o",
            str(output_file),
            "-l",
            prefix,
        ]
    )

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    assert output_file.exists(), f"Output file not created: {output_file}"
    # Optionally, add more assertions to verify the content of the output file


@pytest.mark.parametrize(
    "amplicon_type, unknown_name", [("16s", "16S-U_UNKNOWN")]
)
def test_search_global(
    script_runner, test_data_dir, amplicon_type, unknown_name
):
    """
    Test for running the search_global command.
    """
    # Prepare input and output directories
    output_unoise3_dir = f"output_unoise3_{amplicon_type}"
    input_file = test_data_dir / output_unoise3_dir / "merged.fq.gz"
    db_file = test_data_dir / output_unoise3_dir / "unoise3_zotu.fa"
    output_tsv = test_data_dir / output_unoise3_dir / "unoise3_zotu.biom"

    # Run the script
    ret = script_runner.run(
        [
            "usearch_workflow.py",
            "search_global",
            "-i",
            str(input_file),
            "-d",
            str(db_file),
            "-o",
            str(output_tsv),
            "--unknown_name",
            unknown_name,
        ]
    )

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    assert output_tsv.exists(), f"Output file not created: {output_tsv}"
    # Optionally, add more assertions to verify the content of the output file


@pytest.mark.parametrize("amplicon_type, prefix", [("16s", "16S-U")])
def test_usearch_workflow(script_runner, test_data_dir, amplicon_type, prefix):
    """
    Test script for running two `usearch_workflow.py` commands:
    1. `workflow_per_sample`
    2. `aggregate_samples`
    """

    # Define test variables
    output_unoise3_dir = f"output_unoise3_{amplicon_type}"

    # Prepare input and output directories
    input_file = test_data_dir / output_unoise3_dir / "merged_isolate.fq.gz"
    json_output_file = test_data_dir / output_unoise3_dir / "unoise3_zotu_isolate.json"
    fasta_output_file = test_data_dir / output_unoise3_dir / "unoise3_zotu_isolate.fa"
    biom_output_file = test_data_dir / output_unoise3_dir / "unoise3_zotu_isolate.biom"

    # First command: workflow_per_sample
    cmd_workflow_per_sample = [
        "usearch_workflow.py",
        "workflow_per_sample",
        "-i",
        str(input_file),
        "-o",
        str(json_output_file),
        # "-l",
        # "whatever",
        "--search",
    ]

    # Run the first command
    ret_workflow_per_sample = script_runner.run(cmd_workflow_per_sample)

    # Assertions for the first command
    assert (
        ret_workflow_per_sample.success
    ), f"Script failed with return code {ret_workflow_per_sample.returncode}"
    assert json_output_file.exists(), f"Output file not created: {json_output_file}"

    # Second command: aggregate_samples
    cmd_aggregate_samples = [
        "usearch_workflow.py",
        "aggregate_samples",
        "-i",
        str(json_output_file),
        "-of",
        str(fasta_output_file),
        "-oc",
        str(biom_output_file),
        "-p",
        prefix,
    ]

    # Run the second command
    ret_aggregate_samples = script_runner.run(*cmd_aggregate_samples)

    # Assertions for the second command
    assert (
        ret_aggregate_samples.success
    ), f"Script failed with return code {ret_aggregate_samples.returncode}"
    assert fasta_output_file.exists(), f"Output file not created: {fasta_output_file}"
    assert biom_output_file.exists(), f"Output file not created: {biom_output_file}"



#!/usr/bin/env python3
"""
Figures that have been tested:
- Euclidean distance on log10 relative abundance at sample x genus (or others) level
- Bray-Curtis distance on count at sample x ZOTU (or others) level
"""

import os

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from skbio.diversity import beta_diversity
from skbio.stats.ordination import pcoa
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import seaborn as sns

from masato.get_abundance import get_otu_count, _agg_along_axis


def _load(ab_path: str, metadata_path: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    meta = pd.read_table(metadata_path, index_col=0)
    df = pd.read_table(ab_path, index_col=0).T

    # drop_samples = df.index.str.contains("-V-") | df.index.str.contains("-SNS-")
    # drop_samples = df.index.str.contains("-5N-")
    # df = df.loc[~drop_samples]
    # meta = meta.loc[~drop_samples]

    nonzero = df.sum(axis=1).astype(bool)
    df = df.loc[nonzero]
    meta = meta.loc[nonzero]

    # source = []
    # for i in meta.index:
    #     if "-S-" in i:
    #         if "-OGAmp2-" in i:
    #             source.append("Swab-new")
    #         elif "-OG-" in i:
    #             source.append("Swab-old")
    #         elif "-OGcomb-" in i:
    #             source.append("Swab-comb")
    #         else:
    #             raise ValueError(i)
    #     if "-V-" in i:
    #         source.append("Vortex")
    #     if "-SNS-" in i:
    #         source.append("Swab no spike-in")
    #     if "-R-" in i:
    #         source.append("R2A")
    #     if "-T-" in i:
    #         source.append("TSA")
    # meta["source"] = source

    rep_number = []
    for i in meta.index:
        # if "-S-" in i:
        if i.startswith("B"):
            rep_number.append("AFRL Boneyard")
        elif i.startswith("C"):
            rep_number.append("Cell concentrate")
        else:
            raise ValueError(i)
    meta["Source"] = rep_number
    # meta["rep"] = "every_replication"

    meta["source"] = "every_source"
    return df, meta


def eigsorted(cov: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    vals, vecs = np.linalg.eigh(cov)
    order = vals.argsort()[::-1]
    return vals[order], vecs[:, order]


def plot_dm(
    df_otu_count: pd.DataFrame,
    df_meta: pd.DataFrame,
    fig_path: str | None = None,
    distance: str = "braycurtis",
    title: str | None = None,
    hue: str | None = None,
    style: str | None = None,
    annotate_dots: bool = False,
    plot_ellipses: bool = False,
    fig: matplotlib.figure.Figure | None = None,
    axs: list[matplotlib.axes.Axes] | None = None,
) -> None:
    n_components = min(10, *df_otu_count.shape)
    n_pcs = min(3, n_components)
    if n_components == 1:
        raise ValueError(
            f"Minimum dimensionality of input data is 2, getting {n_components}."
        )

    if distance == "braycurtis":
        bd = beta_diversity("braycurtis", df_otu_count)
        dist_mtx = np.nan_to_num(bd.data, 0)
        pc_obj = pcoa(dist_mtx, number_of_dimensions=n_components)
        pc = pc_obj.samples.copy()
        pc.index = df_otu_count.index
        variance = pc_obj.proportion_explained.to_numpy()
    elif distance == "euclid":
        pc_obj = PCA(n_components=n_components).fit(df_otu_count)
        pc = pc_obj.transform(df_otu_count)
        pc = pd.DataFrame(
            pc,
            index=df_otu_count.index,
            columns=[f"PC{i}" for i in range(1, pc.shape[1] + 1)],
        )
        variance = pc_obj.explained_variance_ratio_
    else:
        raise NotImplementedError(f"Unsupported distance metric: {distance}")

    if n_pcs == 2:
        pc["PC3"] = 0
        variance = np.append(variance, 0)

    marker_size = 8
    pc = pd.concat([pc, df_meta], axis=1)

    # markers = {"Bulk": "X", "Plate-R2A": "o", "Plate-TSA": "s"}
    if hue is None:
        pc["_hue"] = "all"
        hue = "_hue"
    hue_order = sorted(pc[hue].unique().tolist())
    if style is None:
        style_order = None
    else:
        style_order = sorted(pc[style].unique().tolist())

    axis_label_fs = 14
    title_fs = 16
    if fig is None or axs is None:
        fig, axs = plt.subplots(1, 2, figsize=(8, 3.5))
    else:
        if not len(axs) == 2:
            raise ValueError(f"Expected 2 axes, got {len(axs)}.")
    axs: list[matplotlib.axes.Axes]
    sns.scatterplot(
        data=pc,
        x="PC1",
        y="PC2",
        hue=hue,
        hue_order=hue_order,
        style=style,
        # markers=markers,
        style_order=style_order,
        linewidth=0,
        legend=False,
        ax=axs[0],
    )
    axs[0].set_xlabel(f"PC1 ({variance[0] * 100:.2f}%)", fontsize=axis_label_fs)
    axs[0].set_ylabel(f"PC2 ({variance[1] * 100:.2f}%)", fontsize=axis_label_fs)

    sns.scatterplot(
        data=pc,
        x="PC2",
        y="PC3",
        hue=hue,
        hue_order=hue_order,
        style=style,
        # markers=markers,
        style_order=style_order,
        linewidth=0,
        ax=axs[1],
    )
    axs[1].set_xlabel(f"PC2 ({variance[1] * 100:.2f}%)", fontsize=axis_label_fs)
    axs[1].set_ylabel(f"PC3 ({variance[2] * 100:.2f}%)", fontsize=axis_label_fs)

    # set axis to be square
    for ax in axs:
        xleft, xright = ax.get_xlim()
        ybottom, ytop = ax.get_ylim()
        ax.set_aspect(abs((xright - xleft) / (ybottom - ytop)), adjustable="box")

    if hue != "_hue":
        # set legend markers under `hue` to o
        hues = pc[hue].unique().tolist()
        handles, labels = axs[1].get_legend_handles_labels()
        for h, label in zip(handles, labels):
            if label in hues:
                # set marker size to s and marker to o
                h.set_marker("o")
            h.set_markersize(marker_size)
        axs[1].legend(handles, labels)

    axs[1].legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)

    if annotate_dots:
        for i, txt in enumerate(pc.index):
            # annotate with small fontsize
            axs[0].annotate(
                txt, (pc["PC1"].iloc[i], pc["PC2"].iloc[i]), fontsize=6, alpha=0.4
            )
            axs[1].annotate(
                txt, (pc["PC2"].iloc[i], pc["PC3"].iloc[i]), fontsize=6, alpha=0.4
            )

    if plot_ellipses:
        for pc_dims, ax in zip([(1, 2), (2, 3)], axs):
            n = len(pc[hue].unique())
            row_colors = sns.color_palette("tab10", n)
            for group in pc[hue].unique():
                x = pc.loc[pc[hue] == group, f"PC{pc_dims[0]}"]
                y = pc.loc[pc[hue] == group, f"PC{pc_dims[1]}"]
                mean_x, mean_y = x.mean(), y.mean()
                cov = np.cov(x, y)
                lambda_, v = eigsorted(cov)
                theta = np.degrees(np.arctan2(*v[:, 0][::-1]))
                w, h = 2 * 2 * np.sqrt(lambda_)
                ell = Ellipse(
                    xy=(mean_x, mean_y),
                    width=w,
                    height=h,
                    angle=theta,
                    # color=row_colors[pc["sample_group"].unique().tolist().index(group)],
                    alpha=0.2,
                )
                ell.set_facecolor(row_colors[pc[hue].unique().tolist().index(group)])
                ell.set_edgecolor("grey")
                ax.add_artist(ell)

    if title is not None:
        fig.suptitle(title, fontsize=title_fs)
    fig.subplots_adjust(top=0.6)
    if fig_path is not None:
        # fig.tight_layout()
        fig.savefig(fig_path, dpi=300, bbox_inches="tight")
        # also save a pdf file
        fig.savefig(os.path.splitext(fig_path)[0] + ".svg", bbox_inches="tight")
    return fig, axs


def main():
    # if __name__ == "__main__":
    import argparse

    # configure matplotlib PDF saving to use text instead of vector graphics
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["svg.fonttype"] = "none"
    # matplotlib.use("TkAgg")

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--otu_count_tsv",
        help="Path to the OTU count table",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--metadata",
        help="Path to the metadata table",
        required=True,
    )
    parser.add_argument("-t", "--otu_taxonomy_tsv", type=str, required=True)
    parser.add_argument(
        "-f",
        "--fig_dir",
        help="Path to the output figure",
        required=True,
    )
    parser.add_argument(
        "-l",
        "--tax_levels",
        help="Taxonomic levels to plot",
        nargs="+",
        required=True,
    )
    parser.add_argument(
        "-s",
        "--sample_group_key",
        default="sample_group",
        help="Column name in the metadata table to group samples",
    )
    parser.add_argument(
        "-r",
        "--rep_group_key",
        default="rep_group",
        help="Column name in the metadata table to group replicates. If null, plot is "
        "generated without aggregating replicates.",
    )
    parser.add_argument("-sp", "--spikein_taxa_key", type=str, default="spike_in")
    parser.add_argument("-w", "--sample_weight_key", default="sample_weight", type=str)
    # parser.add_argument(
    #     "--relative",
    #     help="Plot relative abundance",
    #     action="store_true",
    # )
    # parser.add_argument(
    #     "--absolute",
    #     help="Plot absolute abundance, exclusive with --relative",
    #     action="store_true",
    # )
    parser.add_argument(
        "-tr",
        "--transform",
        choices=["count", "relative", "absolute"],
        default="count",
        help="Transform to apply to the OTU count table",
    )
    parser.add_argument(
        "-lg",
        "--log10",
        help="Log10 transform",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--distance",
        help="Distance metric",
        required=True,
        choices=["braycurtis", "euclid"],
    )
    parser.add_argument(
        "--hue",
        help="Column name in the metadata table to color samples",
        default=None,
    )
    parser.add_argument(
        "--style",
        help="Column name in the metadata table to style samples",
        default=None,
    )
    parser.add_argument(
        "-a",
        "--annotate_dots",
        help="Annotate dots with sample names",
        action="store_true",
    )
    parser.add_argument(
        "-e",
        "--ellipses",
        help="Plot ellipses",
        action="store_true",
    )
    args = parser.parse_args()

    otu_count_tsv = args.otu_count_tsv
    metadata = args.metadata
    otu_taxonomy_tsv = args.otu_taxonomy_tsv
    fig_dir = args.fig_dir
    tax_levels = args.tax_levels
    transform = args.transform
    sample_group_key = args.sample_group_key
    rep_group_key = args.rep_group_key
    spikein_taxa_key = args.spikein_taxa_key
    sample_weight_key = args.sample_weight_key
    log10 = args.log10
    distance = args.distance
    hue = args.hue
    style = args.style
    annotate_dots = args.annotate_dots
    ellipses = args.ellipses

    df_otu_count, df_meta, df_tax = get_otu_count(
        otu_count_tsv,
        metadata,
        otu_taxonomy_tsv,
        sample_weight_key=sample_weight_key,
        spikein_taxa_key=spikein_taxa_key,
    )

    # [PCA|PCoA (with Bray-Curtis distance)] on[log10 | ]<[tax_level> [relative abundance|absolute abundance|raw count]
    # use keyword format
    title_template = "{distance} on {log10}{tax_level} {transform}"
    title_kw = {}
    if transform == "relative":
        df_otu = df_otu_count.div(df_otu_count.sum(axis=1), axis=0)
        title_kw["transform"] = "relative abundance"
    elif transform == "absolute":
        df_otu = df_otu_count.div(df_meta.norm_factor, axis=0)
        title_kw["transform"] = "absolute abundance"
    elif transform == "count":
        df_otu = df_otu_count.copy()
        title_kw["transform"] = "raw count"
    else:
        raise ValueError(f"Unsupported transform: {transform}")

    if sample_group_key is not None and rep_group_key is not None:
        group = df_meta[sample_group_key] + "-" + df_meta[rep_group_key]
        df_otu = _agg_along_axis(df_otu, group, axis=0)
        df_meta = df_meta.groupby(group).first()
    if log10:
        title_kw["log10"] = "log10 "
        df_otu = np.log10(df_otu + 1e-8)
    else:
        title_kw["log10"] = ""

    if distance == "braycurtis":
        title_kw["distance"] = "PCoA (Bray-Curtis)"
    elif distance == "euclid":
        title_kw["distance"] = "PCA"
    else:
        raise ValueError(
            f"Unsupported distance metric: {distance}, select from 'braycurtis' or 'euclid'."
        )

    os.makedirs(fig_dir, exist_ok=True)
    for level in tax_levels:
        df_otu_tax = _agg_along_axis(df_otu, df_tax[level], axis=1)
        fig_path = os.path.join(
            fig_dir,
            f"{distance}_{'log10_' if log10 else ''}{transform}_{'g' if rep_group_key is not None else 's'}_{level}.png",
        )
        if level == "otu":
            level = "ZOTU"
        else:
            level = level.capitalize()
        plot_dm(
            df_otu_tax,
            df_meta,
            fig_path,
            distance=distance,
            title=title_template.format(**title_kw, tax_level=level),
            hue=hue,
            style=style,
            annotate_dots=annotate_dots,
            plot_ellipses=ellipses,
        )



import subprocess
from pathlib import Path

import pytest
from pytest_console_scripts import ScriptRunner


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


@pytest.fixture
def barcodes_files():
    return {
        "barcodes_fwd": "isolate/barcodes_fwd.fasta",
        "barcodes_rev": "isolate/barcodes_rev.fasta",
        "patterns": "isolate/patterns.txt",
    }


@pytest.mark.parametrize(
    "mode, input_folder, output_file_name, use_barcodes",
    [
        ("isolate_150", "fastq_isolate_16s", "merged_isolate.fq.gz", True),
        ("r1", "fastq_bulk_16s", "merged_bulk.fq.gz", False),
        ("simple", "fastq_bulk_16s", "merged.fq.gz", False),  # Added the simple mode
    ],
)
def test_trim(
    script_runner: ScriptRunner,
    test_data_dir,
    barcodes_files,
    mode,
    input_folder,
    output_file_name,
    use_barcodes,
):
    """
    Generalized test for `trim.py` with three modes: `isolate_150`, `r1`, and `simple`.
    """
    input_dir = test_data_dir / input_folder
    output_file = test_data_dir / "output_unoise3_16s" / output_file_name
    output_file.parent.mkdir(parents=True, exist_ok=True)

    # If the mode is `isolate_150`, we will pass barcodes
    extra_args = []
    if use_barcodes:
        barcodes_fwd = barcodes_files["barcodes_fwd"]
        barcodes_rev = barcodes_files["barcodes_rev"]
        patterns = barcodes_files["patterns"]
        extra_args = ["-fb", barcodes_fwd, "-rb", barcodes_rev, "-pt", patterns]

    # Run the script
    ret = script_runner.run(
        [
            "trim.py",
            "--mode",
            mode,
            "-i",
            str(input_dir),
            "-o",
            str(output_file),
            "-p",
            "16s",
            "-k",
            "110",
            "-l",
            "110",
        ]
        + extra_args
    )

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    assert output_file.exists(), f"Output file not created: {output_file}"
    # Optionally, add more assertions to verify the content of the output file


# def test_merge(test_data_dir):
#     """
#     Merge `merged_bulk.fq.gz` and `merged_isolate.fq.gz` using `zcat` and `gzip`.
#     """
#     bulk_file = test_data_dir / "output_unoise3_16s" / "merged_bulk.fq.gz"
#     isolate_file = test_data_dir / "output_unoise3_16s" / "merged_isolate.fq.gz"
#     output_file = test_data_dir / "output_unoise3_16s" / "merged.fq.gz"

#     command = f"zcat {bulk_file} {isolate_file} | gzip > {output_file}"

#     # Run the shell command
#     proc = subprocess.Popen(
#         command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
#     )
#     stdout, stderr = proc.communicate()

#     # Assertions
#     assert proc.returncode == 0, f"Script failed with return code {proc.returncode}. Error: {stderr.decode()}"
#     assert output_file.exists(), f"Output file not created: {output_file}"



import argparse
import os
import subprocess
import gzip
import shutil
from glob import glob
from collections import Counter
import tempfile
import sys
from contextlib import contextmanager

import pandas as pd
from Bio import Seq, SeqIO
from Bio.SeqRecord import SeqRecord
from umi_tools import UMIClusterer
from tqdm.auto import tqdm
from rich import print as rprint

from masato.trim import (
    get_primer_set,
    get_rc,
    ECREC_DR,
    ECREC_LEADER,
    ECREC_SPACER0,
    RECORDING_PRIMER_3,
    RECORDING_PRIMER_5,
    get_min_overlap,
)
from masato.utils import print_command, smart_open, find_paired_end_files
from masato.read_processer.map_utils import map_se


@contextmanager
def suppress_stdout():
    """
    Context manager to temporarily suppress stdout.
    """
    # Save the original stdout file descriptor
    original_stdout = sys.stdout
    try:
        # Redirect stdout to a null device
        sys.stdout = open(os.devnull, "w")
        yield
    finally:
        # Restore the original stdout
        sys.stdout.close()
        sys.stdout = original_stdout


def fastp_merge(sample: str, read1: str, read2: str, output_dir: str, cpus: int):
    [
        os.makedirs(os.path.join(output_dir, d), exist_ok=True)
        for d in ["merged_direct", "unpaired", "unmerged", "failed", "log"]
    ]

    merge_cmd = [
        "fastp",
        "-i",
        read1,
        "-I",
        read2,
        "-w",
        str(cpus),
        # "--cut_tail",  # don't cut right, since some middle parts seem low quality
        "--merge",
        "--correction",
        "--overlap_len_require",
        "15",
        "--overlap_diff_limit",
        "5",
        "--overlap_diff_percent_limit",
        "20",
        "--merged_out",
        f"{output_dir}/merged_direct/{sample}.fq.gz",
        "--unpaired1",
        f"{output_dir}/unpaired/{sample}.1.fq.gz",
        "--unpaired2",
        f"{output_dir}/unpaired/{sample}.2.fq.gz",
        "--out1",
        f"{output_dir}/unmerged/{sample}.1.fq.gz",
        "--out2",
        f"{output_dir}/unmerged/{sample}.2.fq.gz",
        "--failed_out",
        f"{output_dir}/failed/{sample}.fq.gz",
        "--html",
        f"{output_dir}/log/{sample}.merge.html",
        "--json",
        f"{output_dir}/log/{sample}.merge.json",
    ]

    with open(f"{output_dir}/log/{sample}.merge.err", "w") as err_file:
        print_command(merge_cmd)
        subprocess.run(merge_cmd, stderr=err_file)


def run_cutadapt(
    input_file: str,
    output_file: str,
    *,
    untrimmed_output: str | bool | None = None,
    log_file: str | None = None,
    adapter_5: str | None = None,
    adapter_3: str | None = None,
    e: int | float = 0.2,
    cores: int,
    min_length: int | None = None,
    rc: bool = False,
):
    # makedirs for all output locations
    [
        os.makedirs(os.path.dirname(f), exist_ok=True)
        for f in [output_file, untrimmed_output, log_file]
        if isinstance(f, str)
    ]

    def add_adapter(cmd: list[str]):
        if adapter_5 is not None:
            cmd.extend(["-g", adapter_5])
        if adapter_3 is not None:
            cmd.extend(["-a", adapter_3])

    if rc:
        seqkit_cmd_1 = [
            "seqkit",
            "seq",
            "-r",
            "-p",
            "-t",
            "DNA",
            "--validate-seq",
            input_file,
        ]
        cutadapt_cmd = [
            "cutadapt",
            "-e",
            str(e),
            "--cores",
            str(cores),
        ]
        add_adapter(cutadapt_cmd)
        if untrimmed_output is False:
            cutadapt_cmd.append("--discard-untrimmed")
        elif isinstance(untrimmed_output, str):
            cutadapt_cmd.extend(["--untrimmed-output", untrimmed_output])
        else:
            raise ValueError("untrimmed_output must be a string or False")
        if min_length is not None:
            cutadapt_cmd.extend(["--minimum-length", str(min_length)])
        cutadapt_cmd.append("-")
        seqkit_cmd_2 = [
            "seqkit",
            "seq",
            "-r",
            "-p",
            "-t",
            "DNA",
            "--validate-seq",
            "-",
            "-o",
            output_file,
        ]

        if log_file is not None:
            with open(log_file, "w") as log:
                print_command(seqkit_cmd_1)
                p1 = subprocess.Popen(seqkit_cmd_1, stdout=subprocess.PIPE)
                print_command(cutadapt_cmd)
                p2 = subprocess.Popen(
                    cutadapt_cmd,
                    stdin=p1.stdout,
                    stdout=subprocess.PIPE,
                    stderr=log,
                )
                print_command(seqkit_cmd_2)
                p3 = subprocess.Popen(
                    seqkit_cmd_2, stdin=p2.stdout, stdout=subprocess.PIPE
                )
                _, _ = p3.communicate()
        else:
            print_command(seqkit_cmd_1)
            p1 = subprocess.Popen(seqkit_cmd_1, stdout=subprocess.PIPE)
            print_command(cutadapt_cmd)
            p2 = subprocess.Popen(
                cutadapt_cmd,
                stdin=p1.stdout,
                stdout=subprocess.PIPE,
            )
            print_command(seqkit_cmd_2)
            p3 = subprocess.Popen(seqkit_cmd_2, stdin=p2.stdout, stdout=subprocess.PIPE)
            _, _ = p3.communicate()

    else:
        cutadapt_cmd = [
            "cutadapt",
            "-e",
            str(e),
            "-o",
            output_file,
            "--cores",
            str(cores),
        ]
        add_adapter(cutadapt_cmd)
        if untrimmed_output is False:
            cutadapt_cmd.append("--discard-untrimmed")
        elif isinstance(untrimmed_output, str):
            cutadapt_cmd.extend(["--untrimmed-output", untrimmed_output])
        elif untrimmed_output is not None:
            raise ValueError("untrimmed_output must be a string or False")
        if min_length is not None:
            cutadapt_cmd.extend(["--minimum-length", str(min_length)])
        cutadapt_cmd.append(input_file)

        if log_file is not None:
            with open(log_file, "w") as log:
                print_command(cutadapt_cmd)
                _ = subprocess.run(cutadapt_cmd, stdout=log)
        else:
            print_command(cutadapt_cmd)
            _ = subprocess.run(cutadapt_cmd)


def cutadapt_fix_1(
    input1: str,
    input2: str,
    output1: str,
    output2: str,
    *,
    log_file: str,
    adapter: str,  # 3' adapter of read 1
    e: int | float = 0.15,
    cores: int,
    min_length: int | None = None,
):
    """Some of reads cannot be merged, likely long amplicons. Here we merge them by
    trimming rightmost DR sequence from 3' end in retain mode. Later we will directly
    stitching together the two reads using biopython by reverse complementing read2.

    In this function, we need to use cutadapt in pair mode, but since read 1 adapter is
    rightmost 3', which cutadapt doesn't implement, we need to first reverse complement
    read 1 and then trim from 5', like in the above run_cutadapt function. We therefore
    specify input and output path for read 2 in cutadapt but pass read 1 by piping.

    Read 2 is not modified at all during cutadapt, but we still need to use pair mode to
    drop those read pairs where read 1 is discarded.

    We will also simply discard untrimmed reads here.
    """
    [
        os.makedirs(os.path.dirname(f), exist_ok=True)
        for f in [output1, output2, log_file]
        if f is not None
    ]
    seqkit_cmd_1 = [
        "seqkit",
        "seq",
        "-r",
        "-p",
        "-t",
        "DNA",
        "--validate-seq",
        input1,
    ]
    cutadapt_cmd = [
        "cutadapt",
        "-e",
        str(e),
        "-g",
        adapter,
        "--discard-untrimmed",
        "--action",
        "retain",
        "--cores",
        str(cores),
        "-o",
        "-",
        "-p",
        output2,
    ]
    if min_length is not None:
        cutadapt_cmd.extend(["--minimum-length", str(min_length)])
    cutadapt_cmd.extend(["-", input2])
    seqkit_cmd_2 = [
        "seqkit",
        "seq",
        "-r",
        "-p",
        "-t",
        "DNA",
        "--validate-seq",
        "-",
        "-o",
        output1,
    ]
    with open(log_file, "w") as log:
        print_command(seqkit_cmd_1)
        p1 = subprocess.Popen(seqkit_cmd_1, stdout=subprocess.PIPE)
        print_command(cutadapt_cmd)
        p2 = subprocess.Popen(
            cutadapt_cmd,
            stdin=p1.stdout,
            stdout=subprocess.PIPE,
            stderr=log,
        )
        print_command(seqkit_cmd_2)
        p3 = subprocess.Popen(seqkit_cmd_2, stdin=p2.stdout, stdout=subprocess.PIPE)
        p3.communicate()


def cutadapt_fix_2(
    input1: str,
    input2: str,
    output1: str,
    output2: str,
    *,
    log_file: str | None = None,
    # adapter1: str,  # 3' adapter of read 1
    adapter2: str,  # 3' adapter of read 2
    e: int | float = 0.15,
    cores: int,
    min_length: int | None = None,
):
    [
        os.makedirs(os.path.dirname(f), exist_ok=True)
        for f in [output1, output2, log_file]
        if f is not None
    ]
    cutadapt_cmd = [
        "cutadapt",
        "-e",
        str(e),
        # "-a",
        # adapter1,
        "-A",
        adapter2,
        "--cores",
        str(cores),
        "-o",
        output1,
        "-p",
        output2,
    ]
    if min_length is not None:
        cutadapt_cmd.extend(["--minimum-length", str(min_length)])
    cutadapt_cmd.extend([input1, input2])
    if log_file is not None:
        with open(log_file, "w") as log:
            print_command(cutadapt_cmd)
            subprocess.run(cutadapt_cmd, stdout=log)
    else:
        print_command(cutadapt_cmd)
        subprocess.run(cutadapt_cmd)


def stitch_reads(read1: str, read2: str, output: str):
    os.makedirs(os.path.dirname(output), exist_ok=True)
    r1 = smart_open(read1)
    r2 = smart_open(read2)
    out = smart_open(output, "w")
    for r1_record, r2_record in zip(SeqIO.parse(r1, "fastq"), SeqIO.parse(r2, "fastq")):
        r1_seq = r1_record.seq
        r2_seq = get_rc(r2_record.seq)
        r1_qual = r1_record.letter_annotations["phred_quality"]
        r2_qual = r2_record.letter_annotations["phred_quality"][::-1]
        record = SeqRecord(
            Seq.Seq(r1_seq + r2_seq),
            id=r1_record.id,
            description=r1_record.description,
        )
        record.letter_annotations["phred_quality"] = r1_qual + r2_qual
        SeqIO.write(record, out, "fastq")
    r1.close()
    r2.close()
    out.close()


def cutadapt_remove_nonedited(
    read1: str,
    read2: str,
    output1: str,
    output2: str,
    *,
    log_file: str | None = None,
    adapter: str,
    cpus: int,
):
    cutadapt_cmd = [
        "cutadapt",
        "-a",
        adapter,
        "--discard-trimmed",
        "-o",
        output1,
        "-p",
        output2,
        "--cores",
        str(cpus),
        read1,
        read2,
    ]
    print_command(cutadapt_cmd)
    if log_file is not None:
        with open(log_file, "w") as log:
            subprocess.run(cutadapt_cmd, stdout=log)
    else:
        subprocess.run(cutadapt_cmd)


def makedirs(*directories):
    [os.makedirs(d, exist_ok=True) for d in directories]


def process_one_sample(
    sample: str,
    read1: str,
    read2: str,
    output_dir: str,
    cpus: int,
    concise_mode: bool = False,
):
    merge_dir = os.path.join(output_dir, "merge")
    os.makedirs(merge_dir, exist_ok=True)

    # remove non edited reads
    os.makedirs(os.path.join(merge_dir, "log"), exist_ok=True)

    if concise_mode:
        os.makedirs(os.path.join(merge_dir, "concise"), exist_ok=True)
        no_editing_seq = f"{ECREC_LEADER}{ECREC_DR}{ECREC_SPACER0}"
        read1_out = f"{merge_dir}/concise/{sample}.1.fq.gz"
        read2_out = f"{merge_dir}/concise/{sample}.2.fq.gz"
        cutadapt_remove_nonedited(
            read1,
            read2,
            output1=read1_out,
            output2=read2_out,
            log_file=f"{merge_dir}/log/{sample}.concise.out",
            adapter=f"{no_editing_seq};min_overlap={int(len(no_editing_seq) * 0.8)}",
            cpus=cpus,
        )
        read1, read2 = read1_out, read2_out

    # Merge pairs and remove adapters
    # log_dir = f"{merge_dir}/log"
    # output_md = f"{merge_dir}/merged_direct/{sample}.fq.gz"
    fastp_merge(
        sample,
        read1=read1,
        read2=read2,
        # read1=f"{merge_dir}/concise/{sample}.1.fq.gz",
        # read2=f"{merge_dir}/concise/{sample}.2.fq.gz",
        output_dir=merge_dir,
        cpus=cpus,
    )

    # unmerged fix
    # output_umf_dir = f"{merge_dir}/unmerged_fix"
    # output_pseduo_merge = f"{merge_dir}/merged_pseudo/{sample}.fq.gz"
    # cutadapt_fix_2(
    #     input1=f"{merge_dir}/unmerged/{sample}.1.fq.gz",
    #     input2=f"{merge_dir}/unmerged/{sample}.2.fq.gz",
    #     output1=f"{output_umf_dir}/{sample}.1.fq.gz",
    #     output2=f"{output_umf_dir}/{sample}.2.fq.gz",
    #     log_file=f"{log_dir}/{sample}.fix.out",
    #     # here adapter is non-internal
    #     adapter2=get_rc(ECREC_DR) + f"X;min_overlap=5",
    #     cores=cpus,
    # )
    # stitch_reads(
    #     read1=f"{output_umf_dir}/{sample}.1.fq.gz",
    #     read2=f"{output_umf_dir}/{sample}.2.fq.gz",
    #     output=output_pseduo_merge,
    # )

    # # concat merged and pseudo-merged
    output_merged = f"{merge_dir}/merged_direct/{sample}.fq.gz"
    # os.makedirs(os.path.dirname(output_merged), exist_ok=True)
    # zcat_proc = subprocess.Popen(
    #     ["zcat", output_md, output_pseduo_merge], stdout=subprocess.PIPE
    # )
    # with smart_open(output_merged, "w") as handle:
    #     pigz_proc = subprocess.Popen(
    #         ["pigz", "-p", str(cpus), "-c"], stdin=zcat_proc.stdout, stdout=handle
    #     )
    #     pigz_proc.communicate()

    # clean illumina adapters
    output_rm_adapters = f"{merge_dir}/cleaned/{sample}.fq.gz"
    run_cutadapt(
        input_file=output_merged,
        output_file=output_rm_adapters,
        log_file=f"{merge_dir}/log/{sample}.clean.out",
        adapter_3=f"{RECORDING_PRIMER_5};rightmost;min_overlap={get_min_overlap(RECORDING_PRIMER_5, frac=0.5)}"
        f"...{RECORDING_PRIMER_3};min_overlap={get_min_overlap(RECORDING_PRIMER_3, frac=0.5)}",
        min_length=100,
        cores=cpus,
    )

    # 5' UMI
    umi5_dir = os.path.join(output_dir, "umi5")
    makedirs(umi5_dir)
    output_umi5_rest = f"{umi5_dir}/rest/{sample}.fq.gz"
    output_umi5_umi = f"{umi5_dir}/umi/{sample}.fq.gz"
    adapter_3 = ECREC_LEADER
    adapter_3 = f"{adapter_3};min_overlap={get_min_overlap(adapter_3)}"
    adapter_5 = f"^NNNNNN{ECREC_LEADER}"
    run_cutadapt(
        input_file=output_rm_adapters,
        output_file=output_umi5_umi,
        untrimmed_output=f"{umi5_dir}/umi_untrimmed/{sample}.fq.gz",
        log_file=f"{umi5_dir}/log/{sample}.umi.out",
        adapter_3=adapter_3,
        e=0.3,
        cores=cpus,
    )
    run_cutadapt(
        input_file=output_rm_adapters,
        output_file=output_umi5_rest,
        untrimmed_output=f"{umi5_dir}/rest_untrimmed/{sample}.fq.gz",
        log_file=f"{umi5_dir}/log/{sample}.rest.out",
        adapter_5=adapter_5,
        e=0.3,
        cores=cpus,
    )

    # 3' UMI
    umi3_dir = os.path.join(output_dir, "umi3")
    makedirs(umi3_dir)
    output_umi3_rest = f"{umi3_dir}/rest/{sample}.fq.gz"
    output_umi3_umi = f"{umi3_dir}/umi/{sample}.fq.gz"
    adapter_5 = ECREC_DR + ECREC_SPACER0
    adapter_5 = f"{adapter_5};min_overlap={get_min_overlap(adapter_5)};rightmost"
    adapter_3 = f"{ECREC_DR}{ECREC_SPACER0}NNNNNN$"
    run_cutadapt(
        input_file=output_umi5_rest,
        output_file=output_umi3_umi,
        untrimmed_output=f"{umi3_dir}/umi_untrimmed/{sample}.fq.gz",
        log_file=f"{umi3_dir}/log/{sample}.umi.out",
        adapter_5=adapter_5,
        cores=cpus,
    )
    run_cutadapt(
        input_file=output_umi5_rest,
        output_file=output_umi3_rest,
        untrimmed_output=f"{umi3_dir}/rest_untrimmed/{sample}.fq.gz",
        log_file=f"{umi3_dir}/log/{sample}.rest.out",
        adapter_3=adapter_3,
        cores=cpus,
    )

    # Spacer rounds
    round_num = 1
    current_input = output_umi3_rest
    while True:
        spacer_dir = os.path.join(output_dir, f"spacer{round_num}")
        makedirs(spacer_dir)
        output_spacer_rest = f"{spacer_dir}/rest/{sample}.fq.gz"
        output_spacer_spacer = f"{spacer_dir}/spacer/{sample}.fq.gz"
        run_cutadapt(
            input_file=current_input,
            output_file=output_spacer_spacer,
            untrimmed_output=f"{spacer_dir}/spacer_untrimmed/{sample}.fq.gz",
            log_file=f"{spacer_dir}/log/{sample}.spacer.out",
            adapter_5=ECREC_DR + f";min_overlap={get_min_overlap(ECREC_DR)};rightmost",
            min_length=5,
            cores=cpus,
        )
        run_cutadapt(
            input_file=current_input,
            output_file=output_spacer_rest,
            untrimmed_output=f"{spacer_dir}/rest_untrimmed/{sample}.fq.gz",
            log_file=f"{spacer_dir}/log/{sample}.rest.out",
            adapter_5=get_rc(ECREC_DR) + f";min_overlap={get_min_overlap(ECREC_DR)}",
            min_length=1,
            cores=cpus,
            rc=True,
        )
        # NOTE: Take the common seq between output_spacer_spacer and output_spacer_rest as the
        # final output_spacer_rest, since we only feed to the next round for those sequences that
        # have spacer at this round.
        # We use seqkit for this and follows its document, which prefer grep over common with 2 files.
        # The input for grep is the bigger file (in our case output_spacer_rest) and the pattern file
        # is the smaller file (output_spacer_spacer).
        # save it to a temp fq.gz
        with tempfile.NamedTemporaryFile(suffix=".fq.gz", delete=False) as temp:
            output_spacer_rest_temp = temp.name
            args_seq = ["seqkit", "seq", "--name", "--only-id", output_spacer_spacer]
            args_grep = [
                "seqkit",
                "grep",
                "--pattern-file",
                "-",
                output_spacer_rest,
                "-o",
                output_spacer_rest_temp,
            ]
            print_command(args_seq)
            prog_seq = subprocess.Popen(
                args_seq, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
            )
            print_command(args_grep)
            prog_grep = subprocess.Popen(
                args_grep,
                stdin=prog_seq.stdout,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
            prog_grep.communicate()
            # args = [
            #     "seqkit",
            #     "common",
            #     output_spacer_rest,
            #     output_spacer_spacer,
            #     "-o",
            #     output_spacer_rest_temp,
            # ]
            # print_command(args)
            # subprocess.run(args)
            shutil.move(output_spacer_rest_temp, output_spacer_rest)

        if all_sequences_empty(output_spacer_rest):
            break

        current_input = output_spacer_rest
        round_num += 1

    # concat spacer for later clustering by running seqkit concat <spacer1> <spacer2> ... <spacern> -o <output>
    for k in range(
        1,
        (
            round_num
            if all_sequences_empty(f"{spacer_dir}/spacer/{sample}.fq.gz")
            else round_num + 1
        ),
    ):
        spacer_files = [
            f"{output_dir}/spacer{i}/spacer/{sample}.fq.gz" for i in range(1, k + 1)
        ]
        output_spacer = f"{output_dir}/spacer{k}/spacer_concat/{sample}.fq.gz"
        os.makedirs(os.path.dirname(output_spacer), exist_ok=True)
        if k == 1:
            shutil.copy(spacer_files[0], output_spacer)
        else:
            seqkit_log = f"{output_dir}/spacer{k}/log/{sample}.concat.err"
            seqkit_cmd = ["seqkit", "concat"] + spacer_files + ["-o", output_spacer]
            print_command(seqkit_cmd)
            with open(seqkit_log, "w") as log:
                subprocess.run(seqkit_cmd, stderr=log)

        # run mmseqs2 to cluster concatenated spacer right after like this:
        mmseqs2_prefix = f"{output_dir}/spacer{k}/mmseqs2_clust/{sample}"
        mmseqs2_log = f"{output_dir}/spacer{k}/log/{sample}.clust.err"
        mmseqs2_out = f"{output_dir}/spacer{k}/log/{sample}.clust.out"
        os.makedirs(os.path.dirname(mmseqs2_prefix), exist_ok=True)
        with tempfile.TemporaryDirectory() as temp_dir:
            mmseqs_cmd = [
                "mmseqs",
                "easy-cluster",
                output_spacer,
                mmseqs2_prefix,
                temp_dir,
                "--min-seq-id",
                "0.95",
                "-c",
                "0.8",
                "--cov-mode",
                "1",
                "--remove-tmp-files",
            ]
            print_command(mmseqs_cmd)
            with open(mmseqs2_log, "w") as log, open(mmseqs2_out, "w") as out:
                subprocess.run(mmseqs_cmd, stderr=log, stdout=out)


def all_sequences_empty(fastq_file: str) -> bool:
    """Check if all sequences in a FASTQ file are empty, or if the file itself is empty."""
    with gzip.open(fastq_file, "rt") as handle:
        for record in SeqIO.parse(handle, "fastq"):
            if len(record.seq) > 0:
                return False
    return True


def collect_spacer_info(
    data_dir: str, sample: str, umi_length: int = 6
) -> pd.DataFrame | None:
    """Return a dataframe and a list of spacers.
    The dataframe contains information for each read. Read name being the index, and columns are:
        - umi_5: str
        - umi_3: str
        - num_spacers: int. Number of spacers in the read
        - spacer_idxs: list[int]. List of spacer index of the following list of spacers.
        - centroid: str. The index of the centroid sequence of the read.
    All reads that have either 5' UMI or 3' UMI will be kept, though following analysis will
    focus on reads with both.

    The list of spacers is a list of spacer sequences that's deduplicated. These sequences
    will be later clustered with MMSeqs2 and centroids will be BLASTed to obtain taxonomy
    information.
    """
    umi5_dir = os.path.join(data_dir, "umi5")
    umi3_dir = os.path.join(data_dir, "umi3")
    umi5_path = os.path.join(umi5_dir, f"umi/{sample}.fq.gz")
    umi3_path = os.path.join(umi3_dir, f"umi/{sample}.fq.gz")
    spacer_dirs = glob(os.path.join(data_dir, "spacer*"))
    spacer_orders = [int(i[-1]) for i in spacer_dirs]

    df_data = []
    spacers_set = set()
    spacers = []
    spacer_idx = -1
    # first add 5' umi and 3' umi
    with gzip.open(umi5_path, "rt") as f:
        seq2umi5 = {record.id: str(record.seq) for record in SeqIO.parse(f, "fastq")}
    umi5_df = pd.Series(seq2umi5, name="umi_5").to_frame()
    with gzip.open(umi3_path, "rt") as f:
        seq2umi3 = {record.id: str(record.seq) for record in SeqIO.parse(f, "fastq")}
    umi3_df = pd.Series(seq2umi3, name="umi_3").to_frame()
    df_data = pd.merge(
        umi5_df, umi3_df, how="outer", left_index=True, right_index=True
    ).fillna("")
    if df_data.empty:
        return None

    # remove reads that don't have both 5' and 3' UMI or have empty UMI
    # df_data = df_data.replace("", np.nan).dropna()
    dfs_clust = []
    seqs = {
        i: {"num_spacers": 0, "spacer_idxs": [], "spacer_lens": [], "spacer_qual": 0}
        for i in df_data.index
    }
    for spacer_order in sorted(spacer_orders):
        spacer_dir = os.path.join(data_dir, f"spacer{spacer_order}")
        spacer_path = os.path.join(spacer_dir, "spacer", f"{sample}.fq.gz")
        clust_path = os.path.join(spacer_dir, "mmseqs2_clust", f"{sample}_cluster.tsv")
        if not os.path.exists(spacer_path):
            continue
        with gzip.open(spacer_path, "rt") as f:
            for idx, record in enumerate(SeqIO.parse(f, "fastq")):
                seq = str(record.seq)
                qual = record.letter_annotations["phred_quality"]
                if not seq:  # skip empty sequences
                    continue
                if seq not in spacers_set:
                    spacers.append(seq)
                    spacers_set.add(seq)
                    spacer_idx += 1
                try:
                    seqs[record.id]["num_spacers"] = spacer_order
                    seqs[record.id]["spacer_idxs"].append(spacer_idx)
                    seqs[record.id]["spacer_lens"].append(len(seq))
                    seqs[record.id]["spacer_qual"] += sum(qual)
                except KeyError:
                    print(
                        f"Spacer {spacer_order} of read {record.id} not found in UMI files."
                    )
        if os.path.exists(clust_path):
            dfs_clust.append(pd.read_table(clust_path, names=["centroid", "seq"]))
    df_data = pd.concat([df_data, pd.DataFrame(seqs).transpose()], axis=1)

    for df in dfs_clust:
        df_data.loc[df.seq, "centroid"] = df.centroid.to_numpy()
    df_data["spacer_qual"] = df_data["spacer_qual"] / (
        df_data["spacer_lens"].map(sum) + 1e-8
    )
    df_data["umi"] = df_data.umi_5 + df_data.umi_3
    df_data["good_umi"] = (df_data.umi_5.str.len() == umi_length) & (
        df_data.umi_3.str.len() == umi_length
    )
    return df_data


def correct_umi(df_data: pd.DataFrame) -> dict[str, str]:
    """UMI correction and deduplication with UMI_tools"""
    clusterer = UMIClusterer(cluster_method="directional")
    read2umi_corrected = {}
    for g, df_g in df_data.query("good_umi").groupby("centroid", dropna=False):
        umis = list(map(lambda x: x.encode(), df_g.umi))
        if len(umis) < 2:
            read2umi_corrected[df_g.index[0]] = umis[0].decode()
        else:
            umis = Counter(umis)
            clustered_umis = clusterer(umis, threshold=1)
            umi2umi_corrected = {u: umis[0] for umis in clustered_umis for u in umis}
            for read, umi in zip(df_g.index, df_g.umi, strict=True):
                read2umi_corrected[read] = umi2umi_corrected[umi.encode()].decode()
    return read2umi_corrected


def add_target(df_data: pd.DataFrame, collapse_mode: str | None = None) -> pd.DataFrame:
    """Add target sequence to the dataframe"""
    df_data = df_data.copy()
    df_data["target"] = None
    df_data["size"] = -1
    if collapse_mode is None:
        pass
    else:
        if collapse_mode == "umi":
            groupby = ["num_spacers", "centroid", "umi_corrected"]
        elif collapse_mode == "centroid":
            groupby = ["centroid"]
        else:
            raise ValueError(f"Unknown collapse mode: {collapse_mode}")
        for _, df_g in df_data.groupby(groupby):
            target = df_g.sort_values("spacer_qual", ascending=False).iloc[0].name
            for i in df_g.index:
                df_data.loc[i, "target"] = target
            df_data.loc[target, "size"] = df_g.shape[0]
    return df_data


def agg_arrays(df_data: pd.DataFrame) -> pd.DataFrame:
    return df_data.loc[df_data.index == df_data.target].copy().sort_values("size")


def collect_spacers(
    df_data: pd.DataFrame,
    data_dir: str,
    output_path: str,
    sample: str,
) -> None:
    spacer_orders = [int(i[-1]) for i in glob(os.path.join(data_dir, "spacer*"))]
    # output_path = os.path.join(data_dir, "collect_spacers", "spacer", f"{sample}.fq.gz")
    # reads without spacers are not relevant
    # if collapse_mode == "umi":
    #     # For those that have the same num spacers and umi_corrected and centroid, take
    #     # the one with the highest spacer_qual.
    #     df_data = df_data.reset_index().groupby(["num_spacers", "centroid", "umi_corrected"]).apply(
    #         lambda x: x.sort_values("spacer_qual", ascending=False).iloc[0]
    #     ).set_index("index")
    # elif collapse_mode == "centroid":
    #     # For those that have the centroid, take the one with the highest spacer_qual.
    #     df_data = df_data.reset_index().groupby("centroid").apply(
    #         lambda x: x.sort_values("spacer_qual", ascending=False).iloc[0]
    #     ).set_index("index")
    # else:
    #     if collapse_mode is not None:
    #         raise ValueError(f"Unknown collapse mode: {collapse_mode}")

    good_reads = set(df_data.index)

    with gzip.open(output_path, "wt") as f_out:
        for spacer_order in sorted(spacer_orders):
            spacer_path = os.path.join(
                data_dir, f"spacer{spacer_order}", "spacer", f"{sample}.fq.gz"
            )
            if not os.path.exists(spacer_path):
                continue
            with gzip.open(spacer_path, "rt") as f_in:
                for record in SeqIO.parse(f_in, "fastq"):
                    if record.id in good_reads:
                        size = df_data.loc[record.id, "size"]
                        # add order to the read name as another :<order>
                        record.id = f"{record.id};order={spacer_order};size={size}"
                        SeqIO.write(record, f_out, "fastq")


def get_fastq_length(file_path: str) -> int:
    """
    Calculate the number of sequences in a gzipped FASTQ file.

    Args:
        file_path (str): Path to the gzipped FASTQ file.

    Returns:
        int: Number of sequences in the gzipped FASTQ file.
    """
    # Return 0 if file is empty
    # if os.path.getsize(file_path) == 0:
    #     return 0

    # Use subprocess to count the number of lines in the gzipped file
    if file_path.endswith(".fa.gz") or file_path.endswith(".fasta.gz"):
        prog = "zcat"
        factor = 2
    elif file_path.endswith(".fq.gz") or file_path.endswith(".fastq.gz"):
        prog = "zcat"
        factor = 4
    elif (
        file_path.endswith(".fa")
        or file_path.endswith(".fasta")
        or file_path.endswith(".fna")
    ):
        prog = "cat"
        factor = 2
    elif file_path.endswith(".fq") or file_path.endswith(".fastq"):
        prog = "cat"
        factor = 4
    else:
        raise ValueError(f"Unknown file extension: {file_path}")

    result = subprocess.run(
        # [prog, file_path, "|", "wc", "-l"],
        f"{prog} {file_path} | wc -l",
        capture_output=True,
        text=True,
        shell=True,
        check=True,
    )

    if result.returncode != 0:
        raise RuntimeError(f"Error counting lines in file: {result.stderr}")

    # Each sequence in a FASTQ file spans four lines
    num_lines = int(result.stdout.strip())
    num_sequences = num_lines // factor
    return num_sequences


# if __name__ == "__main__":
def main():
    # data_dir = "/mnt/c/aws_data/20240701_yuanyuan_recording/read_process"
    # fastq_dir = "/mnt/c/aws_data/20240701_yuanyuan_recording/fastq_raw"
    # ref_genome = "/mnt/c/aws_data/20240701_yuanyuan_recording/ref/ecoli_bl21_prec.fna"

    parser = argparse.ArgumentParser(description="Process some directories and files.")
    parser.add_argument(
        "--output_dir", type=str, required=True, help="Path to the data directory"
    )
    parser.add_argument(
        "--fastq_dir", type=str, required=True, help="Path to the fastq directory"
    )
    parser.add_argument(
        "--collapse_mode",
        type=str,
        default=None,
        choices=["umi", "centroid"],
        help="Collapse mode for spacer sequences",
    )
    parser.add_argument(
        "--ref_genome",
        type=str,
        required=True,
        help="Path to the reference genome file",
    )
    parser.add_argument("--quiet", action="store_true", help="Suppress output")

    args = parser.parse_args()

    output_dir = args.output_dir
    fastq_dir = args.fastq_dir
    collapse_mode = args.collapse_mode
    ref_genome = args.ref_genome
    quiet = args.quiet

    if quiet:
        global print_command

        def print_command(*args, **kwargs):
            return None

    bar = tqdm(find_paired_end_files(fastq_dir), leave=False)
    for i, (read1, read2, sample) in enumerate(bar):
        if "Undetermined" in sample:
            continue
        # if i < 66:
        #     continue
        bar.set_description(sample)
        # ==================================
        rprint(f"[bold green]Extracting spacers from reads of {sample}..[/bold green]")
        with suppress_stdout():
            process_one_sample(
                sample,
                # read1=glob(f"{fastq_dir}/{sample}_S*_L001_R1_001.fastq.gz")[0],
                # read2=glob(f"{fastq_dir}/{sample}_S*_L001_R2_001.fastq.gz")[0],
                read1=read1,
                read2=read2,
                output_dir=output_dir,
                cpus=4,
            )

        # # ==================================
        rprint("[bold green]Correcting UMI..[/bold green]")
        spacer_info_path = f"{output_dir}/collect_spacers/spacer_info/{sample}_full.tsv"
        spacer_info_agg_path = f"{output_dir}/collect_spacers/spacer_info/{sample}.tsv"
        os.makedirs(os.path.dirname(spacer_info_path), exist_ok=True)
        df_data = collect_spacer_info(output_dir, sample)
        if df_data is None:
            print(f"WARNING: No arrays with UMI for {sample}, skipping mapping")
            continue
        if df_data.num_spacers.max() == 0:
            print(f"WARNING: No arryas with spacer found in {sample}, skipping mapping")
            df_data.to_csv(spacer_info_path, sep="\t")
            continue
        read2umi_corrected = correct_umi(df_data)
        df_data["umi_corrected"] = df_data.index.map(read2umi_corrected)
        assert df_data.query("good_umi").umi_corrected.notna().all()
        df_data_full = add_target(df_data, collapse_mode)
        df_data_agg = agg_arrays(df_data_full)
        # save spacer info
        df_data_full.to_csv(spacer_info_path, sep="\t")
        df_data_agg.to_csv(spacer_info_agg_path, sep="\t")

        # ==================================
        rprint("[bold green]Collecting spacer sequences..[/bold green]")
        df_data_full = pd.read_table(spacer_info_path, index_col=0)
        df_data_agg = pd.read_table(spacer_info_agg_path, index_col=0)
        output_full_path = f"{output_dir}/collect_spacers/spacer_full/{sample}.fq.gz"
        output_agg_path = f"{output_dir}/collect_spacers/spacer/{sample}.fq.gz"
        os.makedirs(os.path.dirname(output_agg_path), exist_ok=True)
        os.makedirs(os.path.dirname(output_full_path), exist_ok=True)
        collect_spacers(
            df_data=df_data_agg,
            data_dir=output_dir,
            output_path=output_agg_path,
            sample=sample,
        )
        collect_spacers(
            df_data=df_data_full,
            data_dir=output_dir,
            output_path=output_full_path,
            sample=sample,
        )

        # # ==================================
        rprint(
            "[bold green]Mapping reads to genome to distinguish exogenous and "
            "endogenous spacers..[/bold green]"
        )
        makedirs(
            *[
                f"{output_dir}/collect_spacers/{i}"
                for i in ["map", "self", "other", "log"]
            ]
        )
        map_se(
            f"{output_dir}/collect_spacers/spacer/{sample}.fq.gz",
            ref_genome,
            bam_file=f"{output_dir}/collect_spacers/map/{sample}.bam",
            bwa_log=f"{output_dir}/collect_spacers/log/{sample}.bwa.log",
            mapped_reads=f"{output_dir}/collect_spacers/self/{sample}.fq.gz",
            unmapped_reads=f"{output_dir}/collect_spacers/other/{sample}.fq.gz",
            args="relaxed",
        )

        # ==================================
        # ref_genome = (
        #     "/mnt/c/aws_data/20240701_yuanyuan_recording/ref/escherichia_others.fna"
        # )
        # makedirs(
        #     *[
        #         f"{output_dir}/collect_spacers/{i}"
        #         for i in ["map_e", "self_no_e", "other_no_e", "other_is_e", "log"]
        #     ]
        # )
        # map_se(
        #     f"{output_dir}/collect_spacers/other/{sample}.fq.gz",
        #     ref_genome,
        #     bam_file=f"{output_dir}/collect_spacers/map_e/{sample}.bam",
        #     bwa_log=f"{output_dir}/collect_spacers/log/{sample}.bwa_e.log",
        #     mapped_reads=f"{output_dir}/collect_spacers/other_no_e/{sample}.fq.gz",
        #     unmapped_reads=f"{output_dir}/collect_spacers/other_is_e/{sample}.fq.gz",
        # )

        # ==================================
        # rprint("[bold green]BLSATing non-self to plasmid database")
        # plasmid_db = (
        #     "/mnt/c/aws_data/20240701_yuanyuan_recording/db/PlasmidDatabaseJan18.db"
        # )
        # subprocess.run(
        #     [
        #         "seqkit",
        #         "fq2fa",
        #         f"{output_dir}/collect_spacers/other/{sample}.fq.gz",
        #         "-o",
        #         f"{output_dir}/collect_spacers/other/{sample}.fa",
        #     ]
        # )
        # os.makedirs(f"{output_dir}/collect_spacers/other_blastn_plasmid", exist_ok=True)
        # subprocess.run(
        #     [
        #         "blastn",
        #         "-query",
        #         f"{output_dir}/collect_spacers/other/{sample}.fa",
        #         "-db",
        #         plasmid_db,
        #         "-task",
        #         "megablast",
        #         "-word_size",
        #         "10",
        #         "-perc_identity",
        #         "95",
        #         "-dust",
        #         "yes",
        #         "-evalue",
        #         "1e-2",
        #         "-max_target_seqs",
        #         "10000",
        #         "-num_threads",
        #         "16",
        #         "-out",
        #         f"{output_dir}/collect_spacers/other_blastn_plasmid/{sample}.out",
        #         "-outfmt",
        #         "6 qseqid qlen sseqid pident length qstart qend sstart send evalue bitscore slen staxids",
        #     ]
        # )

        # ==================================
        # rprint(
        #     "[bold green]Mapping reads to genome to distinguish exogenous and "
        #     "endogenous spacers..[/bold green]"
        # )
        # ref_genome = (
        #     "/mnt/c/aws_data/20240701_yuanyuan_recording/selected_plasmids/agg.fna"
        # )
        # makedirs(
        #     *[
        #         f"{output_dir}/collect_spacers/{i}"
        #         for i in ["map_plasmid", "mapped_plasmid", "unmapped_plasmid"]
        #     ]
        # )
        # map_se(
        #     f"{output_dir}/collect_spacers/other/{sample}.fq.gz",
        #     ref_genome,
        #     bam_file=f"{output_dir}/collect_spacers/map_plasmid/{sample}.bam",
        #     # bwa_log=f"{output_dir}/collect_spacers/log/{sample}.bwa.log",
        #     mapped_reads=f"{output_dir}/collect_spacers/mapped_plasmid/{sample}.fq.gz",
        #     unmapped_reads=f"{output_dir}/collect_spacers/unmapped_plasmid/{sample}.fq.gz",
        #     args="relaxed",
        # )



import asyncio
from collections import deque
from threading import Event, Thread
import subprocess


class ProcessQueue:
    def __init__(self, capacity):
        self.capacity = capacity
        self.queue = deque()
        self.active = []
        self.all_done_event = Event()

    def add_process(
        self,
        args,
        stdin=None,
        stdout=None,
        stderr=None,
        shell=False,
        cwd=None,
        env=None,
    ) -> None:
        self.add_callable(
            subprocess.Popen,
            args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            shell=shell,
            cwd=cwd,
            env=env,
        )

    def add_callable(self, callable_, *args, **kwargs) -> None:
        """Add a callable that starts a process."""
        if not self.queue and len(self.active) == 0:
            self.all_done_event.clear()
        self.queue.append([callable_, args, kwargs])
        self._try_to_start_process()

    def _try_to_start_process(self):
        while self.queue and len(self.active) < self.capacity:
            callable_, args, kwargs = self.queue.popleft()
            process = callable_(*args, **kwargs)
            self.active.append(process)
            thread = Thread(target=self._monitor_process, args=(process,))
            thread.start()

    def _monitor_process(self, process):
        process.wait()
        self.active.remove(process)
        if not self.active and not self.queue:
            # Set the event if there are no active processes and nothing in the queue
            self.all_done_event.set()
        else:
            self._try_to_start_process()

    def wait_for_completion(self):
        """Wait until all tasks have completed."""
        self.all_done_event.wait()


class AsyncProcessQueue:
    def __init__(self, capacity):
        self.capacity = capacity
        self.queue = deque()
        self.active = set()  # Using a set to track active tasks
        self.all_done_event = asyncio.Event()
        self.all_done_event.set()  # Initialize as set indicating no tasks are pending

    async def add_process(
        self,
        args,
        stdin=None,
        stdout=None,
        stderr=None,
        shell=False,
        cwd=None,
        env=None,
    ) -> None:
        # Wrap process creation in a callable for consistent handling
        process_callable = lambda: (
            asyncio.create_subprocess_exec(
                *args, stdin=stdin, stdout=stdout, stderr=stderr, cwd=cwd, env=env
            )
            if not shell
            else asyncio.create_subprocess_shell(
                args, stdin=stdin, stdout=stdout, stderr=stderr, cwd=cwd, env=env
            )
        )
        await self.add_callable(process_callable)

    async def add_callable(self, callable_):
        """Add a callable that starts an async process."""
        if not self.queue and not self.active:
            self.all_done_event.clear()
        self.queue.append(callable_)
        await self._try_to_start_process()

    async def _try_to_start_process(self):
        while self.queue and len(self.active) < self.capacity:
            callable_ = self.queue.popleft()
            task = asyncio.create_task(callable_())
            self.active.add(task)
            task.add_done_callback(self._task_done)

    def _task_done(self, task):
        self.active.remove(task)
        if not self.active and not self.queue:
            self.all_done_event.set()  # No more active tasks
        else:
            asyncio.create_task(
                self._try_to_start_process()
            )  # Try to start more tasks if capacity allows

    async def wait_for_completion(self):
        """Wait until all tasks have completed."""
        await self.all_done_event.wait()


async def main():
    queue = AsyncProcessQueue(capacity=2)
    commands = [
        'echo "Task 1 started"; sleep 1; echo "Task 1 completed"',
        'echo "Task 2 started"; sleep 2; echo "Task 2 completed"',
        'echo "Task 3 started"; sleep 3; echo "Task 3 completed"',
        'echo "Task 4 started"; sleep 2; echo "Task 4 completed"',
        'echo "Task 5 started"; sleep 1; echo "Task 5 completed"',
    ]
    for cmd in commands:
        await queue.add_process(cmd, shell=True)
    await queue.wait_for_completion()
    print("All processes have completed.")


if __name__ == "__main__":
    asyncio.run(main())
    # queue = ProcessQueue(capacity=2)
    # commands = [
    #     'echo "Task 1 started"; sleep 1; echo "Task 1 completed"',
    #     'echo "Task 2 started"; sleep 2; echo "Task 2 completed"',
    #     'echo "Task 3 started"; sleep 3; echo "Task 3 completed"',
    #     'echo "Task 4 started"; sleep 2; echo "Task 4 completed"',
    #     'echo "Task 5 started"; sleep 1; echo "Task 5 completed"',
    # ]
    # for cmd in commands:
    #     queue.add_process(cmd, shell=True)
    # queue.wait_for_completion()
    # print("All processes have completed.")
